import nidaqmx
import nidaqmx.constants as niconst

import measpy.signal as ms

import numpy as np
from numpy.matlib import repmat

from datetime import datetime

def n_to_ain(n):
    return 'ai'+str(n-1)
def n_to_aon(n):
    return 'ao'+str(n-1)


def callback(task_handle, every_n_samples_event_type,
                number_of_samples, callback_data):
    print('Every N Samples callback invoked.')

    return 0


def ni_run_measurement(M):
    system = nidaqmx.system.System.local()
    nsamps = int(round(M.dur*M.fs))

    if M.device_type!='ni':
        print("Warning: deviceType != 'ni'. Changing to 'ni'.")
        M.device_type='ni'
    if M.in_device=='':
        print("Warning: no output device specified, changing to "+system.devices[0].name)
        M.in_device=system.devices[0].name

    if hasattr(M, 'in_range'):
        if M.in_range == None:
            inr = False
        else :
            inr = True
    else:
        inr = False
    if not(inr):
        val = nidaqmx.system.device.Device(M.in_device).ai_voltage_rngs[-1]
        print("Warning: no input range specified, changing to the max value of "+M.in_device+" -> "+str(val))
        M.in_range = list(val for b in M.in_map)

    if M.out_sig!=None:
        if M.out_device=='':
            print("Warning: no output device specified, changing to "+system.devices[0].name)
            M.out_device=system.devices[0].name
        if hasattr(M, 'out_range'):
            if M.out_range == None:
                outr = False
            else :
                outr = True
        else:
            outr = False
        if not(outr):
            val = nidaqmx.system.device.Device(M.out_device).ao_voltage_rngs[-1]
            print("Warning: no output range specified, changing to the max value of "+M.out_device+" -> "+str(val))
            M.out_range = list(val for b in M.out_map)
            
    now = datetime.now()
    M.date = now.strftime("%Y-%m-%d")
    M.time = now.strftime("%H:%M:%S")

    # Insert a synchronization peak at the begining of the output signals
    if M.out_sig==None:
        dursync=0
        effsync=False
    elif M.io_sync>0:
        if M.io_sync in M.in_map:
            nout = M.x.shape[1]
            peaks = repmat(ms.picv(M.fs),nout,1).T
            zers = repmat(np.zeros(int(M.fs)),nout,1).T
            outx = np.block([[peaks],[M.x],[zers]])
            effsync = True
            dursync=4
            indsearch=M.in_map.index(M.io_sync)
        else:
            print('io_sync channel not present in in_map, no sync is done')
            outx=M.x
            dursync=0
            effsync=False
    else:
        outx=M.x
        dursync=0
        effsync=False

    nsamps = int(round((dursync+M.dur)*M.fs))

    intask = nidaqmx.Task(new_task_name="in") # read task

    if M.out_sig!=None:
        outtask = nidaqmx.Task(new_task_name="out") # write task

    # Set up the read tasks
    for i,n in enumerate(M.in_map):
        print(n_to_ain(n))
        intask.ai_channels.add_ai_voltage_chan(
            physical_channel=M.in_device + "/" + n_to_ain(n),
            terminal_config=niconst.TerminalConfiguration.DEFAULT,
            min_val=-M.in_range[i], max_val=M.in_range[i],
            units=niconst.VoltageUnits.VOLTS)

    intask.timing.cfg_samp_clk_timing(
        rate=M.fs,
        sample_mode=niconst.AcquisitionType.CONTINUOUS,
        samps_per_chan=nsamps)

    if M.out_sig!=None:
        # Set up the write tasks, use the sample clock of the Analog input if possible
        for i,n in enumerate(M.out_map):   
            outtask.ao_channels.add_ao_voltage_chan(
                physical_channel=M.out_device + "/" + n_to_aon(n), 
                min_val=-M.out_range[i], max_val=M.out_range[i],
                units=niconst.VoltageUnits.VOLTS)
      
        if M.in_device.startswith('myDAQ'):
            # If the device is a myDAQ card, we keep most default values
            # The myDAQ devices are set up separately because
            # there is no error messages when setting up properties
            # that are not supported, and the acquisition then fails
            print("This is a NI my DAQ device")
            outtask.timing.cfg_samp_clk_timing(
                rate=M.fs,
                sample_mode=niconst.AcquisitionType.CONTINUOUS,
                samps_per_chan=nsamps)
        else:
            try:
                # We first try to use analog input sample clock as output clock
                outtask.timing.cfg_samp_clk_timing(
                    rate=M.fs,
                    source="/" + M.in_device + "/ai/SampleClock", #"OnboardClock",
                    sample_mode=niconst.AcquisitionType.CONTINUOUS,
                    samps_per_chan=nsamps)
                print('Use of /'+ M.in_device + '/ai/SampleClock as output clock : success !')
            except:
                # If it fails, use defaults
                # Then the in/out are not synchronized
                # There is hence the possibility to use one analog input
                # to do the in/out sync (io_sync=input channel number)
                print("Error when choosing \""+"/" + M.in_device + "/ai/SampleClock\" as clock source, let's try \"OnboardClock\" ")
                outtask.timing.cfg_samp_clk_timing(
                    rate=M.fs,
                    sample_mode=niconst.AcquisitionType.CONTINUOUS,
                    samps_per_chan=nsamps)

        if len(M.out_map)==1:
            outtask.write(outx[:,0], auto_start=False)
        else:
            outtask.write(outx.T, auto_start=False)

        outtask.start() # Start the write task first, waiting for the analog input sample clock

    y = intask.read(nsamps,timeout=M.dur+10) # Start the read task

    intask.close()

    if M.out_sig!=None:
        outtask.close()

    y=np.array(y).T

    if effsync:
        posmax = int( np.argmax(y[int(0.25*M.fs*2):int(0.75*M.fs*2),indsearch]) + 0.75*M.fs*2 )
        print(posmax)
        y = y[posmax:posmax+M.fs*M.dur,:]

    if len(M.in_map)==1:
        M.data[M.in_name[0]].raw = y
    else:
        n=0
        for s in M.in_name: 
            M.data[s].raw = y[:,n]
            n+=1

def ni_get_devices():
    system = nidaqmx.system.System.local()
    print(system.devices)
    return system.devices



# measurement.py
# 
# A class for measurement management with data acquisition devices
#
# OD - 2021

import measpy.signal as ms
from measpy.signal import Signal

from measpy._tools import csv_to_dict, convl, convl1

import numpy as np
import matplotlib.pyplot as plt

from copy import copy

import scipy.io.wavfile as wav
import csv
import pickle
import json

from unyt import Unit

class Measurement:
    """ The class Measurement allows to simply define and perform
        a measurement.

        Initialization parameters:

        :param fs: Sampling frequency, defaults to 44100.
        :type fs: int
        :param dur: Duration in seconds, defaults to 2.0.
        :type dur: float
        :param in_map: Map of the inputs used on the device, defaults to [1,2].
        :type in_map: list(int)
        :param in_cal: Calibrations of the incoming signals, in volts/units, defaults to a list of 1.0.
        :type in_cal: list(str)
        :param in_unit: Measurement units, defaults to a list of 'V'.
        :type in_unit: list(str)
        :param in_dbfs: Input voltage for a unitary value in the resulting signal, defaults to 1.0.
        :type in_unit: list(str)
        :param in_name: Short name of the measurement signals (used as keys for the data dictionnary), defaults to 'In'+str(n)
        :type in_name: list(str)
        :param in_desc: Description of the measurement signals, defaults to 'Thi is input '+str(n)
        :type in_unit: list(str)
        :param out_sig: Output signal type ('noise', 'logsweep', 'file.wav' or None) where file.wav is a wave file string.
        :type out_sig: str
        :param extrat: (not really useful yet) Additionnal extra time before and after (additionnal time), in sample numbers.
        :type extrat: tuple(int,int)

        :param out_map: Map of the outputs used on the device, defaults to [1]
        :type out_map: list(int)
        :param out_amp: Output amplitude, defaults to 1.0
        :type out_amp: float
        :param out_dbfs: Output voltage for a unitary value in the sent signal, defaults to 1.0.
        :type out_sig: float
        :param extrat: Additionnal extra time before and after (additionnal time), in sample numbers.
        :type extrat: tuple(int,int)
        :param out_sig_freqs: Minimum and maximum frequencies for the generated output signals (used if out_sig is 'noise' or 'logsweep')
        :type out_sig_freqs: tuple(float,float)
        :param fades: Fades in/out at the begining and end of the output signal, defaults to (0,0)
        :type fades: tuple(int,int)
        :param io_sync: (not implemented yet) Specifies if in/out synchronization is done, and which type, defaults to 0 (no synchronization).
        :type io_sync: int

        :param device_type: Type of device 'audio', 'ni', 'pico' or '', or None, defaults to ''. It can be eventually reactualized when running the measurement.
        :type device_type: str or None
        :param in_device: Input device, defaults to ''
        :type in_device: str
        :param out_device: Output device, defaults to ''
        :type out_device: str

    """
    def __init__(self, **params):
        # params checking
        if 'out_sig' in params:
            noise=params['out_sig']!='noise'
            logsweep=params['out_sig']!='logsweep'
            sine=params['out_sig']!='sine'
            wa=not str(params['out_sig']).upper().endswith('.WAV')
            non=params['out_sig']!=None
            ar=type(params['out_sig'])!=np.ndarray
            if noise&logsweep&sine&wa&non&ar:
                raise Exception("out_sig must but be numpy.ndarray, 'noise', 'sweep', 'sine', '*.wav' or None")

        self.fs = params.setdefault("fs",44100)
        self.dur = params.setdefault("dur",2.0)
        self.device_type = params.setdefault("device_type",'')

        if 'in_name' in params:
            self.in_map = params.setdefault("in_map",list(range(1,len(params['in_name'])+1)))
        else:
            self.in_map = params.setdefault("in_map",[1,2])
        self.in_name = params.setdefault("in_name",list('In'+str(b) for b in self.in_map))

        self.in_map = params.setdefault("in_map",[1,2])
        self.in_cal = params.setdefault("in_cal",list(1.0 for b in self.in_map))
        self.in_dbfs = params.setdefault("in_dbfs",list(1.0 for b in self.in_map))
        self.in_unit = params.setdefault("in_unit",list('V' for b in self.in_map))
        self.in_name = params.setdefault("in_name",list('In'+str(b) for b in self.in_map))
        self.in_desc = params.setdefault("in_desc",list('This is input '+str(b) for b in self.in_map))
        self.out_sig = params.setdefault("out_sig",'noise')                
        self.extrat = params.setdefault("extrat",[0.0,0.0])
        self.in_device = params.setdefault("in_device",'')
        if self.out_sig!=None:
            self.out_map = params.setdefault("out_map",[1])
            self.out_amp =  params.setdefault("out_amp",1.0)
            self.out_dbfs = params.setdefault("out_dbfs",list(1.0 for b in self.out_map))
            self.out_name = params.setdefault("out_name",list('Out'+str(b) for b in self.out_map))
            self.out_desc =  params.setdefault("out_desc",list('This is output '+str(b) for b in self.out_map))
            self.out_sig_freqs =  params.setdefault("out_sig_freqs",[20.0,20000.0])
            self.io_sync = params.setdefault("io_sync",0)
            self.out_sig_fades = params.setdefault("out_sig_fades",[0,0])
            self.out_device = params.setdefault("out_device",'')
        self.data = {}
        if self.out_sig!=None:
            for n in range(len(self.out_name)):
                self.data[self.out_name[n]]=Signal(desc=self.out_desc[n],
                                                    fs=self.fs,
                                                    unit='V',
                                                    cal=1.0,
                                                    dbfs=self.out_dbfs[n])
            self.create_output()
        for n in range(len(self.in_name)):
            self.data[self.in_name[n]]=Signal(desc=self.in_desc[n],
                                                fs=self.fs,
                                                unit=self.in_unit[n],
                                                cal=self.in_cal[n],
                                                dbfs=self.in_dbfs[n])
        self.data_keys = list(self.data.keys())
        if self.device_type=='pico':
            self.in_range = params.setdefault("in_range",list('10V' for b in self.in_map))
            self.upsampling_factor = params.setdefault("upsampling_factor",1)
            self.in_coupling = params.setdefault("in_coupling",list('dc' for b in self.in_map))
        if self.device_type=='ni':
            self.in_range = params.setdefault("in_range",None)
            self.out_range = params.setdefault("out_range",None)

    def create_output(self):
        """ Creates the output signals, if out_sig is 'noise',
            'logsweep' or a string ending with 'wav'.
            If 'out_sig' is None, nothing is created.
        """
        if self.out_sig=='noise': # White noise output signal
            self.data[self.out_name[0]] = self.data[self.out_name[0]].similar(
                volts=ms._noise(self.fs,self.dur,self.out_amp,self.out_sig_freqs)
            ).fade(self.out_sig_fades).add_silence(self.extrat)

            if self.out_map==0:
                self._out_map=[1]

        elif self.out_sig=='logsweep': # Logarithmic sweep output signal
            self.data[self.out_name[0]] = self.data[self.out_name[0]].similar(
                volts=ms._log_sweep(self.fs,self.dur,self.out_amp,self.out_sig_freqs)
            ).fade(self.out_sig_fades).add_silence(self.extrat)

            if self.out_map==0:
                self.out_map=[1]

        elif self.out_sig=='sine':  # Sinusoidal output signal
            self.data[self.out_name[0]] = self.data[self.out_name[0]].similar(
                volts=ms._sine(self.fs,self.dur,self.out_amp,self.out_sig_freqs[0])
            ).fade(self.out_sig_fades).add_silence(self.extrat)
        
            if self.out_map==0:
                self.out_map=[1]

        elif self.out_sig.upper().endswith('.WAV'): # Wave file output signal

            rate, x = wav.read(self.out_sig)

            if len(x.shape)==1:
                nchan = 1
                x=x[:,None]
            else:
                nchan = x.shape[1]

            if (rate!=self.fs):
                print('Warning: ')
                print('  Rate of input file='+str(rate)+'Hz.')
                print("  It's different than fs="+str(self.fs)+"Hz of current measurement object.")
                print('  Changing to: '+str(rate))
                self.fs=rate
            if (x.shape[0]/self.fs!=self.dur):
                print("Warning:")
                print("  Duration of input file="+str(x.shape[0]/self.fs)+"s.")
                print("  It's different than dur="+str(self.dur)+"s of current measurement object.")
                print('  Changing to: '+str(x.shape[0]/self.fs))
                self.dur = x.shape[0]/self.fs
            if self.out_map==0:
                self.out_map=list(range(1,x.shape[1]+1))
            elif (nchan!=len(self.out_map)):
                print("Warning:")
                print("Size of out_map and number of channels do not correspond.")
                if (x.shape[1]<len(self.out_map)):
                    print("  Truncating current out_map...")
                    self.out_map=self.out_map[0:nchan]
                    self.out_name=self.out_name[0:nchan]
                    self.out_dbfs=self.out_dbfs[0:nchan]
                else:
                    print("  Truncating channels of the output signal...")
                    x=x[:,0:len(self.out_map)]

            if x.dtype == 'int16':
                vmax=32768
            elif x.dtype == 'int32':
                vmax=2147483648
            else:
                vmax=1.0
                
            for ii in range(len(self.out_map)):
                self.data[self.out_name[ii]]=self.data[self.out_name[ii]].similar(
                    volts=np.array(x[:,ii],dtype=float)/vmax
                ).fade(self.out_sig_fades).add_silence(self.extrat)
        elif type(self.outsig)==np.ndarray:
            pass

    def show(self):
        """ Pretty prints the measurement properties """
        print("Measurement with the following properties:")
        print("| Type of device: device_type="+self.device_type)
        print("| Input device: device="+str(self.in_device))
        print("| Sampling frequency (Hz): fs="+str(self.fs))
        print("| Duration (s): dur="+str(self.dur))
        st = str(self.in_map)
        print("| Input map: in_map="+st)
        st = str(self.in_cal)
        print("| Input calibrations (V/Unit): in_cal="+st)
        st = str(self.in_dbfs)
        print("| Input 0dBFS (V): in_dbfs="+st)
        st = "', '".join(self.in_unit)
        print('| Input units: in_unit='+"['"+st+"']")
        st = "', '".join(self.in_name)
        print('| Input names: in_name='+"['"+st+"']")        
        st = "', '".join(self.in_desc)
        print('| Input descriptions: in_desc='+"['"+st+"']")        
        print('| Extra time before and after: extrat='+str(self.extrat))
        if self.out_sig!=None:
            print("| Output device: device="+str(self.out_device))
            st = str(self.out_map)
            print('| Output map: out_map='+st)
            st = "', '".join(self.out_name)
            print('| Output names: out_name='+"['"+st+"']")              
            print('| Output amp: out_amp='+str(self.out_amp))
            print("| Output signal type: out_sig='"+self.out_sig+"'")
            print('| Min and max frequency of generated output: out_sig_freqs='+str(self.out_sig_freqs))
            print('| Synchonisation: ioSync='+str(self.io_sync))
            st = str(self.out_dbfs)
            print("| Output 0dBFS (V): out_dbfs="+st)
        try:
            print("| Measurement date: date='"+self.date+"'")
            print("| Measurement time: time='"+self.time+"'")
        except:
            print("| No measurement date or time, the measurement hasn't been performed")

        if self.device_type=='pico':
            print('| Input ranges: in_range='+str(self.in_range))
            print("| Upsampling factor: upsampling_factor='"+str(self.upsampling_factor)+"'")
            print("| Input coupling: in_coupling="+str(self.in_coupling))
        print("[ Contents of the dictionnary data (keys):")
        for key in self.data:
            print("| "+key)

    def __repr__(self):
        out = "measpy.Measurement("
        out += "fs="+str(self.fs)
        out += ", dur="+str(self.dur)
        out += ", device_type='"+str(self.device_type)+"'"
        out += ", in_device='"+str(self.in_device)+"'"
        out += ', in_name='+str(self.in_name)
        out += ', in_desc='+str(self.in_desc)
        out += ', in_map='+str(self.in_map)
        out += ', in_cal='+str(self.in_cal)
        out += ', in_unit='+str(self.in_unit)
        out += ', in_dbfs='+str(self.in_dbfs)
        out += ', extrat='+str(self.extrat)
        try:
            out += ", date='"+self.date+"'"
            out += ", time='"+self.time+"'"
        except:
            pass
        if self.out_sig!=None:
            out += ", out_device='"+str(self.out_device)+"'"
            out += ', out_name='+str(self.out_name)
            out += ', out_desc='+str(self.out_desc)
            out += ', out_map='+str(self.out_map)
            out += ', out_dbfs='+str(self.out_dbfs)
            out += ", out_sig='"+str(self.out_sig)+"'"
            out += ', out_sig_freqs='+str(self.out_sig_freqs)
            out += ', out_sig_fades='+str(self.out_sig_fades)
            out += ", out_amp="+str(self.out_amp)
            out += ", io_sync="+str(self.io_sync)
        if self.device_type=='pico':
            out += ", in_range="+str(self.in_range)
            out += ", upsampling_factor="+str(self.upsampling_factor)
            out += ", in_coupling="+str(self.in_coupling)
        out +=")"
        
        return out

    def _to_dict(self,withdata=True):
        """ Converts a Measurement object to a dict

            :param withdata: Optionnally removes the data arrays, defaults to True
            :type withdata: bool
            
        """
        self.data_keys = list(self.data.keys())
        mesu = copy(self.__dict__)
        if not(withdata):
            del mesu['data']
        return mesu

    def _from_dict(self,meas):
        """ Update Measurement properties from a dict,
            generally loaded from a file.

            :param meas: dictionnary whose contents should be compatible
            with Measurement class
        """
        self.fs=convl1(float,meas['fs'])
        self.dur=convl1(float,meas['dur'])
        self.in_map=convl(int,meas['in_map'])
        self.in_cal=convl(float,meas['in_cal'])
        self.in_dbfs=convl(float,meas['in_dbfs'])
        self.in_unit=convl(str,meas['in_unit'])
        self.in_name=convl(str,meas['in_name'])
        self.in_desc=convl(str,meas['in_desc'])
        self.extrat=convl(float,meas['extrat'])
        self.out_sig=convl1(str,meas['out_sig'])
        if self.out_sig=='None':
            self.out_sig=None
        try:
            self.data=meas['data']
        except:
            pass
        try:
            self.date=convl1(str,meas['date'])
            self.time=convl1(str,meas['time'])
        except:
            pass
        if self.out_sig!=None:
            self.out_map=convl(int,meas['out_map'])
            self.out_amp=convl1(float,meas['out_amp'])
            self.out_name=convl(str,meas['out_name'])
            self.out_desc=convl(str,meas['out_desc'])
            self.out_sig_freqs=convl(float,meas['out_sig_freqs'])
            self.out_sig_fades=convl(float,meas['out_sig_fades'])
            self.io_sync=convl1(int,meas['io_sync'])
            self.out_dbfs=convl(float,meas['out_dbfs'])
            self.out_device=convl1(str,meas['out_device'])            
            self.out_device=convl1(str,meas['out_device'])
        self.in_device=convl1(str,meas['in_device'])
        self.device_type=convl1(str,meas['device_type'])
        self.data_keys=convl(str,meas['data_keys'])
        if self.device_type=='pico':
            self.in_range=convl(str,meas['in_range'])
            self.upsampling_factor=convl1(int,meas['upsampling_factor'])            
            self.in_range=convl(str,meas['in_coupling'])
        
        # print('In _from_dict')
        # print(self.data)

    def to_pickle(self,filename):
        with open(filename, 'wb') as handle:
            pickle.dump(self._to_dict(), handle, protocol=pickle.HIGHEST_PROTOCOL)

    @classmethod
    def from_pickle(cls,filename):
        """ Load a Measurement class object from a pickle file

            :param filename: Filename
            :type filename: str

            :return: The loaded measurement object
            :rtype: Measurement

        """
        with open(filename, 'rb') as handle:
            mesu = pickle.load(handle)
        M = cls()
        M._from_dict(mesu)
        for key in M.datakeys:
            M.data[key].unit=Unit(str(M.data[key].unit))
        return M

    def _data_to_array_old(self,includetime=False,datatype='raw'):
        n = 0
        if includetime:
            out = self.data[list(self.data.keys())[0]].time[:,None]
            n += 1
        for key in self.data.keys():
            if n==0:
                if datatype=='raw':
                    out = self.data[key].raw[:,None]
                elif datatype=='volts':
                    out = self.data[key].volts[:,None]
                elif datatype=='values':
                    out = self.data[key].values[:,None]
                n += 1
            else:
                if datatype=='raw':
                    out = np.block([out,self.data[key].raw[:,None]])
                elif datatype=='volts':
                    out = np.block([out,self.data[key].volts[:,None]])
                elif datatype=='values':
                    out = np.block([out,self.data[key].values[:,None]])
                n += 1
        return out

    def _data_to_array(self,includetime=False,datatype='raw'):
        if self.out_sig!=None:
            if datatype=='raw':
                outdata = np.concatenate((self.x_raw,self.y_raw),1)
            elif datatype=='volts':
                outdata = np.concatenate((self.x_volts,self.y_volts),1)
            elif datatype=='values':
                outdata = np.concatenate((self.x,self.y),1)
        else:
            if datatype=='raw':
                outdata = self.y_raw
            elif datatype=='volts':
                outdata = self.y_volts
            elif datatype=='values':
                outdata = self.y
        if includetime:
            outdata = np.concatenate((self.t[:,None],outdata),1)
        return outdata

    def _data_to_wav(self,filename):
        """ Save all data in the measurement as a unique wav file
            It is not the recommended usage to use this method, but
            use to_csvwav or to_jsonwav to save the data, and its
            descriptions.

            :param filename: WAV file name
            :type filename: str
        """
        wav.write(filename,int(round(self.fs)),self._data_to_array())

    def _data_to_txt(self,filename,includetime=True,datatype='raw'):
        """ Save all data in the measurement as a unique txt file
            It is not the recommended usage to use this method, but
            use to_csvtxt or to_jsontxt to save the data, and its
            descriptions.

            :param filename: TXT file name
            :type filename: str
            :param datatype: Type of data ('raw', 'volts' or 'values')
            :type datatype: str 
        """
        np.savetxt(filename,self._data_to_array(includetime=includetime,datatype=datatype))


    def _data_from_wav(self,filename):
        _, dat = wav.read(filename)
        n = 0
        if len(self.data_keys)==1:
            for key in self.data_keys:
                self.data[key].raw = dat
        else:
            for key in self.data_keys:
                self.data[key].raw = dat[:,n]
                n += 1        #print(self.data)
        for key in self.in_name:
            pos = self.in_name.index(key)
            self.data[key].unit = Unit(self.in_unit[pos])
            self.data[key].cal = self.in_cal[pos]
            self.data[key].dbfs = self.in_dbfs[pos]
            self.data[key].desc = self.in_desc[pos]
            self.data[key].fs = self.fs
        if self.out_sig!=None:
            for key in self.out_name:
                pos = self.out_name.index(key)
                self.data[key].unit = Unit('V')
                self.data[key].dbfs = self.out_dbfs[pos]
                self.data[key].desc = self.out_desc[pos]
                self.data[key].fs = self.fs

    def _params_to_csv(self,filename):
        """ Writes all the Measurement object parameters to a csv file """
        dd = self._to_dict(withdata=False)
        #data_keys = list(self.data.keys())
        with open(filename, 'w', newline='') as file:
            writer = csv.writer(file)
            for key in dd:
                if type(dd[key])==list:
                    writer.writerow([key]+dd[key])
                else:
                    writer.writerow([key,str(dd[key])])
            #writer.writerow(['data_keys']+data_keys)
                    
    def _csv_to_params(self,filename):
        """ Load measurement parameters from a csv file """
        self._from_dict(csv_to_dict(filename))

    def _params_to_json(self,filename):
        """ Writes all the Measurement object parameters to a json file """
        with open(filename, mode='w', encoding='utf-8') as f:
            json.dump(self._to_dict(withdata=False), f, indent=2)

    def _json_to_params(self,filename):
        """ Load measurement parameters from a json file """
        with open(filename, encoding='utf-8') as f:
            self._from_dict(json.load(f))

    def to_csvwav(self,filebase):
        """ Saves a Measurement object to a set of files

            * filebase : string from which two file names are created
            * filebase+'.csv' : All measurement parameters
            * filebase+'.wav' : all input and out channels + time (32 bit float WAV at fs)
        """
        self._params_to_csv(filebase+'.csv')
        try:
            self._data_to_wav(filebase+'.wav')
        except:
            print('data_to_wav failed (no data?)')

    @classmethod
    def from_csvwav(cls,filebase):
        """ Load a measurement object from a set of files

            * filebase : string from which two file names are created
            * filebase+'.csv' : All measurement parameters
            * filebase+'.wav' : all input and out channels + time (32 bit float WAV at fs)
        """
        M=cls()
        M._csv_to_params(filebase+'.csv')
        if M.out_sig!=None:
            M=cls(out_name=M.out_name,in_name=M.in_name)
        else:
            M=cls(in_name=M.in_name)
        M._csv_to_params(filebase+'.csv')
        #try:
        M._data_from_wav(filebase+'.wav')
        #except:
        #    print(filebase+'.wav')
        #    print('data_from_wav failed (file not present?)')
        return M

    def to_jsonwav(self,filebase):
        """ Saves a Measurement object to a set of files

            * filebase : string from which two file names are created
            * filebase+'.json' : All measurement parameters
            * filebase+'.wav' : all input and out channels + time (32 bit float WAV at fs)
        """
        self._params_to_json(filebase+'.json')
        try:
            self._data_to_wav(filebase+'.wav')
        except:
            print('data_to_wav failed (no data?)')

    @classmethod
    def from_jsonwav(cls,filebase):
        """ Load a measurement object from a set of files
        
            * filebase : string from which two file names are created
            * filebase+'.json' : All measurement parameters
            * filebase+'.wav' : all input and out channels + time (32 bit float WAV at fs)
        """
        M=cls()
        M._json_to_params(filebase+'.json')
        if M.out_sig!=None:
            M=cls(out_name=M.out_name,in_name=M.in_name)
        else:
            M=cls(in_name=M.in_name)
        M._json_to_params(filebase+'.json')
        # try:
        M._data_from_wav(filebase+'.wav')
        # except:
        # print('data_from_wav failed (file not present?)')
        return M

    def to_csvtxt(self,filebase,includetime=True,datatype='values'):
        """ Saves a Measurement object to a set of files

            * filebase : string from which two file names are created
            * filebase+'.csv' : All measurement parameters
            * filebase+'.txt' : all input and out channels + time (as readable textfile)
        """
        self._params_to_csv(filebase+'.csv')
        # try:
        self._data_to_txt(filebase+'.txt',includetime=includetime,datatype=datatype)
        # except:
        #     print('data_to_wav failed (no data?)')

    def plot(self,ytype='values',limit=None):

        for ii in range(self.y.shape[1]):
            if ytype=='values':
                plt.plot(self.t,self.y[:,ii])
            if ytype=='volts':
                plt.plot(self.t,self.y_volts[:,ii])
            if ytype=='raw':
                plt.plot(self.t,self.y_raw[:,ii])
        if limit!=None:
            plt.plot(self.t,limit*np.ones_like(self.t),':',color='grey')
            plt.plot(self.t,-1*limit*np.ones_like(self.t),':',color='grey')
        plt.xlabel('Time(s)')
        legende = []
        for ii in range(self.y.shape[1]):
            if ytype=='values':
                legende+=[self.in_name[ii]+'('+self.in_unit[ii]+')']
            if ytype=='volts':
                legende+=[self.in_name[ii]+'(volts)']
            if ytype=='raw':
                legende+=[self.in_name[ii]+'(-)']            
        legende+=['limits']
        plt.legend(legende)
        plt.title('Measurement date: '+str(self.date)+"   "+str(self.time))
        #plt.grid('on',color='grey',linestyle=':')
        plt.grid('on')

    def tfe(self,nperseg=2**16,noverlap=None,plotH=False):
        """ Helper function that calculates the transfer function between
            the output channel x and all the input channels y. Works only
            if x has only one channel.
            If out_sig='logsweep', the method of Farina is used, Welch's
            method is used otherwise.

            Retruns: a dict of Spectral class objects
        """
        if (self.out_sig=='noise') or (self.out_sig.upper().endswith('.WAV')):
            if self.x.shape[1]>1:
                print("tfe : This basic helper function works only if out_sig has only one channel")
                return None
            out={}
            for key in self.in_name:
                out[key]=self.data[key].tfe_welch(self.data[self.out_name[0]],nperseg=nperseg,noverlap=noverlap)                
        elif self.out_sig=='logsweep':
            if self.x.shape[1]>1:
                print("tfe : This basic helper function works only if out_sig has only one channel")
                return None
            out = {}
            for key in self.in_name:
                out[key]=self.data[key].tfe_farina(self.out_sig_freqs)
        else:
            print("tfe : This basic helper function works only if ouSig='noise' or 'logsweep'")
            return None, None
        if plotH:
            for key in self.in_name:
                out[key].plot()
        return out

    @property
    def x(self):
        """ The output data values as a 2D-array converted in the correponding
            units using the calibration and dbfs properties
        """
        return np.array([self.data[n].values for n in self.out_name]).T
    
    @property
    def y(self):
        """ The input data values as a 2D-array converted in the correponding
            units using the calibration and dbfs properties
        """
        return np.array([self.data[n].values for n in self.in_name]).T

    @property
    def x_volts(self):
        """ The output data values as a 2D-array converted in the correponding
            units using the calibration and dbfs properties
        """
        return np.array([self.data[n].volts for n in self.out_name]).T
    
    @property
    def y_volts(self):
        """ The input data values as a 2D-array converted in the correponding
            units using the calibration and dbfs properties
        """
        return np.array([self.data[n].volts for n in self.in_name]).T

    @property
    def x_raw(self):
        """ Raw output values as a 2D-array (no conversion applied) """
        return np.array([self.data[n].raw for n in self.out_name]).T
    
    @property
    def y_raw(self):
        """ Raw input values as a 2D-array (no conversion applied) """
        return np.array([self.data[n].raw for n in self.in_name]).T

    @property
    def t(self):
        """ Time array """
        return ms.create_time(self.fs,dur=self.dur+self.extrat[0]+self.extrat[1])

    # Old tfe function (deprecated)
    def tfeb(self,nperseg=2**16,noverlap=None,plotH=False):
        """ DEPRECATED
            Helper function that calculates the transfer function between
            the output channel x and all the input channels y. Works only
            if x has only one channel.
            If out_sig='logsweep', the method of Farina is used, Welch's
            method is used otherwise.

            Returns : 1D f array and 2D H array 
        """
        print('Warning: DEPRECATED')
        if (self.out_sig=='noise') or (self.out_sig.upper().endswith('.WAV')):
            if self.x.shape[1]>1:
                print("tfe : This basic helper function works only if out_sig has only one channel")
                return None, None
            Hout = np.zeros((1+int(nperseg/2),self.y.shape[1]),dtype=complex)
            for ii in range(self.y.shape[1]):
                freqs, Hout[:,ii] =  ms.tfe_welch(self.x[:,0],
                                    self.y[:,ii],
                                    nperseg=nperseg,
                                    noverlap=noverlap,
                                    fs=self.fs)
        elif self.out_sig=='logsweep':
            #H = np.zeros_like(self.y,dtype=complex)
            freqs, Hout = ms.tfe_farina(self.y[:,0],
                                self.fs,
                                self.out_sig_freqs)
            Hout = Hout[:,None]
            for ii in range(self.y.shape[1]-1):
                freqs, H =  ms.tfe_farina(self.y[:,ii+1],
                                    self.fs,
                                    self.out_sig_freqs)
                Hout = np.block([Hout,H[:,None]])
        else:
            print("tfe : This basic helper function works only if ouSig='noise' or 'logsweep'")
            return None, None
        if plotH:
            ms.plot_tfe(freqs,Hout)
        return freqs, Hout

    def tfe_xy(self,x,y,plotH=False,**kwargs):
        """ DEPRECATED
            Compute transfert function between x and y, where x and y are
            strings representing keys of the dictionnary of the data property
            of the Measurement object. Welch's method is used. Data is calibrated.
        """
        print('Warning: DEPRECATED')
        out = self.data[y].tfe_welch(self.data[x],**kwargs)

        if plotH:
            out.plot()
        return out
        



# __main__.py

import matplotlib.pyplot as plt
import numpy as np
import sys
import measpy as mp

def main():
    if len(sys.argv)>1:
        if sys.argv[1]=='audio':
            from measpy.audio import audio_run_measurement as run_measurement
        elif sys.argv[1]=='ni':
            from measpy.ni import ni_run_measurement as run_measurement
        elif sys.argv[1]=='ps2000':
            from measpy.pico import ps2000_run_measurement as run_measurement
        elif sys.argv[1]=='ps4000':
            from measpy.pico import ps4000_run_measurement as run_measurement
    M1 = mp.Measurement(out_sig=None,
                    in_map=[1,2],
                    in_desc=['In1','In2'],
                    in_cal=[1.0,1.0],
                    in_unit=['V','V'],
                    in_dbfs=[1.0,1.0],
                    extrat=[0,0],
                    dur=2)
    run_measurement(M1)
    M1.plot()
    M1.to_csvwav(str(sys.argv[3]))

if __name__ == '__main__':
    main()



import unittest
import sys
sys.path.insert(0, "measpy")
import measpy as mp
import numpy as np
import unyt
import matplotlib.pyplot as plt
from os import remove

err_tol = 1e-4
class TestSignal(unittest.TestCase):

    global s1, sp1, s2, s3, s4, s5, dur, fs
    dur = 3.0
    fs = 44100

    # A 3s signal with a even number of samples
    s1 = mp.Signal.noise(fs=fs,dur=dur,amp=1.0,freqs=[20.0,20000.0],unit='1',cal=1.0,dbfs=1.0)
    
    # A 3s signal with an odd number of samples
    sp1 = mp.Signal.noise(fs=fs+1,dur=dur,amp=1.0,freqs=[20.0,20000.0],unit='1',cal=1.0,dbfs=1.0)
    s3 = mp.Signal.sine(fs=44100,freq=441,dur=1.0)
    s4 = mp.Signal.log_sweep(fs=44100,freqs=[20,20000],dur=1.0)
    s5 = mp.Signal.log_sweep(fs=48000,freqs=[20,20000],dur=1.0)

    def test_durations(self):
        for s in (s1,sp1):
            self.assertEqual(s.dur,dur,"Calculation in Signal.dur is wrong ?")
            self.assertEqual(1/(s.time[1]-s.time[0]),s.fs,"Calculation of Signal.time is wrong ?")
            self.assertEqual(s.length,s.fs*s.dur,"Calculation of Signal.length is wrong ?")
            
    def test_signalops(self):
        self.assertLess((s1-s1).max(),err_tol, "Difference of two equal signals should be zero" )
        self.assertLess(((s1/s1)-1).max(),err_tol, "Difference of two equal signals should be zero" )
        self.assertLess( abs((abs(-(s1/s1))-1).max()), err_tol, "Absolute value of a -1 constant signal should be one"  )
        self.assertLess( abs((abs(-(s1*(1/s1)))-1).max()), err_tol, "Absolute value of a -1 constant signal should be one"  )

    def test_corr(self):
        self.assertLess(abs(abs(s1.corr(s1.similar(values=np.roll(s1.values,s1.fs))).tmax())-1*unyt.s), err_tol, "Max of correlation between signals should be at 1 sec.")
        self.assertLess(abs(abs(s1.add_silence(extrat=(1,0)).corr(s1.add_silence(extrat=(0,1))).tmax())-1*unyt.s), err_tol, "Max of correlation between signals should be at 1 sec.")
        self.assertLess(abs(abs((s1.timelag(s1.similar(values=np.roll(s1.values,s1.fs)))))-1), err_tol, "Calculated timelag should be 1 sec.")

    def test_normalize(self):
        self.assertEqual(s1.normalize().max(), 1.0, "Should be one")

    def test_rms(self):
        self.assertLess(s1.rms()-np.sqrt(2), err_tol, "RMS of 1V sine should be sqrt(2)")

    def test_hilbert(self):
        self.assertLess((s3.timelag(s3.hilbert())-1/(4*s3.fs)),err_tol, "Hilbert transform should induce a pi/2 (1/4 period) phase shift")

    def test_fft(self):
        for s in (s1,sp1):
            print (s)
            self.assertEqual(s.fft().dur,dur,"Calculation of Spectral.dur is wrong ?")
            self.assertEqual(s.rfft().dur,dur,"Calculation of Spectral.dur is wrong ?")
            self.assertEqual(s.fft().freqs[1],1/s.dur,"Calculation of frequency vector is wrong ?")
            self.assertLess(min(abs(s.rfft(norm='ortho').values_at_freqs(range(30,19000,100))))-1,err_tol,"Noise spectrum has to be unitary in the (20,20000) frequency range")
            self.assertLess((s.rfft().irfft()-s).rms(),err_tol,"RFFT + IRFFT gets the same signal")
            self.assertLess((s.fft().ifft()-s).rms(),err_tol,"FFT + IFFT gets the same signal")
            self.assertEqual(s.fft().sample_number,s.length,"After FFT the sample number of the original signal is correct")
            self.assertEqual(s.rfft().sample_number,s.length,"After RFFT the sample number of the original signal is correct")
            
    def test_resample(self):
        a1=abs(s4.rfft(norm='forward').values_at_freqs(range(30,19000,100)))
        a2=abs(s5.rfft(norm='forward').values_at_freqs(range(30,19000,100)))
        self.assertLess(max(abs(a2-a1)),err_tol, "Difference between signals should be zero")

        a1=abs(s4.rfft(norm='forward').values_at_freqs(range(30,19000,100)))
        a2=abs(s5.resample(s4.fs).rfft(norm='forward').values_at_freqs(range(30,19000,100)))
        self.assertLess(max(abs(a2-a1)),err_tol, "Difference between signals should be zero")

    def test_importexport(self):
        filname = 'test'+str(int(np.random.random()*10000))
        s1.to_csvwav(filname)
        self.assertLess((s1-mp.Signal.from_csvwav(filname)).max(),err_tol, "Difference of two equal signals should be zero" )
        remove(filname+".csv")
        remove(filname+".wav")

if __name__ == '__main__':
    unittest.main()



# __init__.py

from measpy.signal import (Signal, 
                            Spectral,
                            Weighting,
                            PREF,
                            DBUREF,
                            DBVREF,
                            WDBA,
                            WDBC)
from measpy.measurement import Measurement



import measpy.signal as ms

import numpy as np

from datetime import datetime
from scipy.signal import decimate

import ctypes
import numpy as np
from picosdk.ps4000 import ps4000
from picosdk.functions import adc2mV, assert_pico_ok

from picosdk.ps2000 import ps2000
from picosdk.functions import assert_pico2000_ok
from picosdk.ctypes_wrapper import C_CALLBACK_FUNCTION_FACTORY
from picosdk.PicoDeviceEnums import picoEnum

import matplotlib.pyplot as plt
from matplotlib.widgets import Button
plt.style.use('seaborn-v0_8')

def findindex(l,e):
    try:
        a=l.index(e)
    except:
        a=None
    return a

def ps2000_run_measurement(M):
    """
    This function needs M to contain the following properties:
        - in_range : a list of strings specifying the voltage range.
            Possible voltage ranges are "10MV", "20MV", "50MV", "100MV",
            "200MV", "500MV", "1V", "2V", "5V", "10V", "20V", "50V", "100V"
        - upsampling_factor : upsampling factor
        - in_coupling : Coupling configuration of the channels.
            Can be "ac" or "dc"
    """

    import time

    CALLBACK = C_CALLBACK_FUNCTION_FACTORY(
        None,
        ctypes.POINTER(ctypes.POINTER(ctypes.c_int16)),
        ctypes.c_int16,
        ctypes.c_uint32,
        ctypes.c_int16,
        ctypes.c_int16,
        ctypes.c_uint32
    )

    # Effective sampling frequency
    # If upsampling_factor is > 1, the actual data acquisition is
    # performed at fs*upsampling_factor, and then decimated to
    # the desired frequency

    effective_fs = M.fs * M.upsampling_factor
    duree_ns = M.dur*1e9

    # Sample interval is the duration between two consecutive samples
    # As it is the sampling frequency that is specified,
    # and the sampling interval can only take integer increments
    # of 1 nanoseconds, the actual sampling frequency might
    # necessitate adjustments

    si = round(1e9/effective_fs)
    if effective_fs != (1e9/si):
        effective_fs = (1e9/si)
        print("Warning : Sampling frequency fs changed to nearest possible value of "+str(effective_fs)+" Hz")
    
    print("Effective sampling frequency: "+str(effective_fs))

    sampleInterval = int(si)

    # Setup channel A
    indA = findindex(M.in_map,1)
    if indA!=None:
        enabledA = True
        rangeA = ps2000.PS2000_VOLTAGE_RANGE['PS2000_'+M.in_range[indA]]
        adc_valuesA = []
        print('Channel A: enabled with range '+'PS2000_'+M.in_range[indA]+' ('+str(rangeA)+')')
        if M.in_coupling[indA].capitalize()=='Dc':
           couplingA = 'PICO_DC'
        elif M.in_coupling[indA].capitalize()=='Ac':
            couplingA = 'PICO_AC'
        else:
            print("Input A coupling not recognized, set to 'dc'")
            couplingA = 'PICO_DC'
            M.in_coupling[indA]='dc'
    else:
        enabledA = False
        rangeA = ps2000.PS2000_VOLTAGE_RANGE['PS2000_10V']
        print('Channel A: disabled')

    # Setup channel B
    indB = findindex(M.in_map,2)
    if indB!=None:
        enabledB = True
        rangeB = ps2000.PS2000_VOLTAGE_RANGE['PS2000_'+M.in_range[indB]]
        adc_valuesB = []
        if M.in_coupling[indB].capitalize()=='Dc':
           couplingB = 'PICO_DC'
        elif M.in_coupling[indB].capitalize()=='Ac':
            couplingB = 'PICO_AC'
        else:
            print("Input B coupling not recognized, set to 'dc'")
            M.in_coupling[indB]='dc'
            couplingB = 'PICO_DC'
        print('Channel B: enabled with range '+'PS2000_'+M.in_range[indB]+' ('+str(rangeB)+')')
    else:
        enabledB = False
        rangeB = ps2000.PS2000_VOLTAGE_RANGE['PS2000_10V']
        couplingB = 'PICO_DC'
        print('Channel B: disabled')

    def get_overview_buffers(buffers, _overflow, _triggered_at, _triggered, _auto_stop, n_values):
        if enabledA:
            adc_valuesA.extend(buffers[0][0:n_values])
            if enabledB:
                adc_valuesB.extend(buffers[2][0:n_values])
        else:
            if enabledB:
                adc_valuesB.extend(buffers[2][0:n_values])

    callback = CALLBACK(get_overview_buffers)

    def adc_to_mv(values, range_, bitness=16):
        v_ranges = [10, 20, 50, 100, 200, 500, 1_000, 2_000, 5_000, 10_000, 20_000]
        return [(x * v_ranges[range_]) / (2**(bitness - 1) - 1) for x in values]

    with ps2000.open_unit() as device:
        print('Device info: {}'.format(device.info))

        res = ps2000.ps2000_set_channel(
            device.handle,
            picoEnum.PICO_CHANNEL['PICO_CHANNEL_A'],
            enabledA,
            picoEnum.PICO_COUPLING[couplingA],
            rangeA,
        )
        res = ps2000.ps2000_set_channel(
            device.handle,
            picoEnum.PICO_CHANNEL['PICO_CHANNEL_B'],
            enabledB,
            picoEnum.PICO_COUPLING[couplingB],
            rangeB,
        )
        assert_pico2000_ok(res)

        res = ps2000.ps2000_run_streaming_ns(
            device.handle,
            sampleInterval,
            2,
            100_000,
            False,
            1,
            20_000
        )
        assert_pico2000_ok(res)

        now = datetime.now()
        M.date = now.strftime("%Y-%m-%d")
        M.time = now.strftime("%H:%M:%S")

        print('Start')
        start_time = time.time_ns()

        while time.time_ns() - start_time < duree_ns*1.1:
            ps2000.ps2000_get_streaming_last_values(
                device.handle,
                callback
            )

        print('Done')

        if M.fs!=effective_fs/M.upsampling_factor:
            M.fs = effective_fs/M.upsampling_factor
            print('Warning : Sampling frequency fs changed to nearest possible value of '+str(M.fs)+' Hz')
            for i in range(len(M.in_map)):
                M.data[M.in_name[i]].fs = M.fs

        for i in range(len(M.in_map)):
            if M.in_map[i] == 1:
                if M.upsampling_factor > 1:
                    M.data[M.in_name[i]].raw = decimate(np.double(adc_to_mv(adc_valuesA, rangeA)[:])/1000,M.upsampling_factor,ftype='fir')[0:int(round(M.dur*M.fs))]
                else:
                    M.data[M.in_name[i]].raw = (np.double(adc_to_mv(adc_valuesA, rangeA)[:])/1000)[0:int(round(M.dur*M.fs))]
            elif M.in_map[i] == 2:
                if M.upsampling_factor > 1:
                    M.data[M.in_name[i]].raw = decimate(np.double(adc_to_mv(adc_valuesB, rangeB)[:])/1000,M.upsampling_factor,ftype='fir')[0:int(round(M.dur*M.fs))]
                else:
                    M.data[M.in_name[i]].raw = (np.double(adc_to_mv(adc_valuesB, rangeB)[:])/1000)[0:int(round(M.dur*M.fs))]

def ps2000_plot(M,plotbuffersize=2000,updatetime=0.1):
    """
    Realtime plot of signal and spectrum

    This function needs M to contain the following properties:
        - in_range : a list of strings specifying the voltage range.
            Possible voltage ranges are "10MV", "20MV", "50MV", "100MV",
            "200MV", "500MV", "1V", "2V", "5V", "10V", "20V", "50V", "100V"
        - upsampling_factor : upsampling factor
        - in_coupling : Coupling configuration of the channels.
            Can be "ac" or "dc"
    """

    import time

    global plotbuffer, ax, x, line, stop, timesincelastupdate, updtime

    updtime = updatetime

    timesincelastupdate = 0.0

    stop = False

    def fstop(event):
        global stop
        stop = True

    def tamp_plus(event):
        ax[0].set_ylim(np.array(ax[0].get_ylim())/2)
    def tamp_moins(event):
        ax[0].set_ylim(np.array(ax[0].get_ylim())*2)
    def famp_plus(event):
        ax[1].set_ylim(np.array(ax[1].get_ylim())/2)
    def famp_moins(event):
        ax[1].set_ylim(np.array(ax[1].get_ylim())*2)
    def on_close(event):
        global stop
        stop = True

    x = np.arange(plotbuffersize)
    plotbuffer = np.sin(x)
    fig, ax = plt.subplots(1,2,figsize=(15,5))
    fig.subplots_adjust(bottom=0.2)
    ax[0].set_xlabel('Temps [s]',fontsize=15)
    ax[0].set_ylabel('Tension [V]',fontsize=15)
    ax[1].set_xlabel('Fréquence [Hz]',fontsize=15)
    ax[1].set_ylabel('Tension [V]',fontsize=15)
    ax[1].set_xlim([0.01,M.fs/2])
    ax[1].set_ylim([0,10])
    linet, = ax[0].plot(x/M.fs,plotbuffer)
    linef, = ax[1].plot(np.fft.fftfreq(n=plotbuffersize,d=1/M.fs),np.fft.fft(plotbuffer,norm='ortho'))
    axs = fig.add_axes([0.4, 0.01, 0.2, 0.075])
    bstop = Button(axs, 'Stop')
    bstop.on_clicked(fstop)
    atxp = fig.add_axes([0.02, 0.6, 0.04, 0.05])
    btplus = Button(atxp, '+')
    btplus.on_clicked(tamp_plus)
    atxm = fig.add_axes([0.02, 0.4, 0.04, 0.05])
    btmoins = Button(atxm, '-')
    btmoins.on_clicked(tamp_moins)
    afxp = fig.add_axes([0.92, 0.6, 0.04, 0.05])
    bfplus = Button(afxp, '+')
    bfplus.on_clicked(famp_plus)
    afxm = fig.add_axes([0.92, 0.4, 0.04, 0.05])
    bfmoins = Button(afxm, '-')
    bfmoins.on_clicked(famp_moins)
    fig.canvas.mpl_connect('close_event', on_close)
    plt.pause(0.0001)

    CALLBACK = C_CALLBACK_FUNCTION_FACTORY(
        None,
        ctypes.POINTER(ctypes.POINTER(ctypes.c_int16)),
        ctypes.c_int16,
        ctypes.c_uint32,
        ctypes.c_int16,
        ctypes.c_int16,
        ctypes.c_uint32
    )

    # Effective sampling frequency
    # If upsampling_factor is > 1, the actual data acquisition is
    # performed at fs*upsampling_factor, and then decimated to
    # the desired frequency

    effective_fs = M.fs * M.upsampling_factor
    duree_ns = M.dur*1e9

    # Sample interval is the duration between two consecutive samples
    # As it is the sampling frequency that is specified,
    # and the sampling interval can only take integer increments
    # of 1 nanoseconds, the actual sampling frequency might
    # necessitate adjustments

    si = round(1e9/effective_fs)
    if effective_fs != (1e9/si):
        effective_fs = (1e9/si)
        print("Warning : Sampling frequency fs changed to nearest possible value of "+str(effective_fs)+" Hz")
    
    print("Effective sampling frequency: "+str(effective_fs))

    sampleInterval = int(si)

    # Setup channel A
    indA = findindex(M.in_map,1)
    if indA!=None:
        enabledA = True
        rangeA = ps2000.PS2000_VOLTAGE_RANGE['PS2000_'+M.in_range[indA]]
        adc_valuesA = []
        print('Channel A: enabled with range '+'PS2000_'+M.in_range[indA]+' ('+str(rangeA)+')')
        if M.in_coupling[indA].capitalize()=='Dc':
           couplingA = 'PICO_DC'
        elif M.in_coupling[indA].capitalize()=='Ac':
            couplingA = 'PICO_AC'
        else:
            print("Input A coupling not recognized, set to 'dc'")
            couplingA = 'PICO_DC'
            M.in_coupling[indA]='dc'
    else:
        enabledA = False
        rangeA = ps2000.PS2000_VOLTAGE_RANGE['PS2000_10V']
        print('Channel A: disabled')

    # Setup channel B
    indB = findindex(M.in_map,2)
    if indB!=None:
        enabledB = True
        rangeB = ps2000.PS2000_VOLTAGE_RANGE['PS2000_'+M.in_range[indB]]
        adc_valuesB = []
        if M.in_coupling[indB].capitalize()=='Dc':
           couplingB = 'PICO_DC'
        elif M.in_coupling[indB].capitalize()=='Ac':
            couplingB = 'PICO_AC'
        else:
            print("Input B coupling not recognized, set to 'dc'")
            M.in_coupling[indB]='dc'
            couplingB = 'PICO_DC'
        print('Channel B: enabled with range '+'PS2000_'+M.in_range[indB]+' ('+str(rangeB)+')')
    else:
        enabledB = False
        rangeB = ps2000.PS2000_VOLTAGE_RANGE['PS2000_10V']
        couplingB = 'PICO_DC'
        print('Channel B: disabled')

    def get_overview_buffers(buffers, _overflow, _triggered_at, _triggered, _auto_stop, n_values):
        global plotbuffer, timesincelastupdate, updtime
        timesincelastupdate += n_values/M.fs
        plotbuffer=np.roll(plotbuffer,int(-n_values),axis=0)
        plotbuffer[-n_values:] = (np.array(adc_to_mv(buffers[0][0:n_values],rangeA))/1000)[:]
        if timesincelastupdate > updtime:
            linet.set_ydata(plotbuffer)
            linef.set_ydata(np.abs(np.fft.fft(plotbuffer*np.hanning(plotbuffersize),norm='ortho')))
            plt.pause(0.0001)
            timesincelastupdate = 0.0
        # if enabledA:
        #     adc_valuesA.extend(buffers[0][0:n_values])
        #     if enabledB:
        #         adc_valuesB.extend(buffers[2][0:n_values])
        # else:
        #     if enabledB:
        #         adc_valuesB.extend(buffers[2][0:n_values])

    callback = CALLBACK(get_overview_buffers)

    def adc_to_mv(values, range_, bitness=16):
        v_ranges = [10, 20, 50, 100, 200, 500, 1_000, 2_000, 5_000, 10_000, 20_000]
        return [(x * v_ranges[range_]) / (2**(bitness - 1) - 1) for x in values]

    with ps2000.open_unit() as device:
        print('Device info: {}'.format(device.info))

        res = ps2000.ps2000_set_channel(
            device.handle,
            picoEnum.PICO_CHANNEL['PICO_CHANNEL_A'],
            enabledA,
            picoEnum.PICO_COUPLING[couplingA],
            rangeA,
        )
        res = ps2000.ps2000_set_channel(
            device.handle,
            picoEnum.PICO_CHANNEL['PICO_CHANNEL_B'],
            enabledB,
            picoEnum.PICO_COUPLING[couplingB],
            rangeB,
        )
        assert_pico2000_ok(res)

        res = ps2000.ps2000_run_streaming_ns(
            device.handle,
            sampleInterval,
            2,
            100_000,
            False,
            1,
            20_000
        )
        assert_pico2000_ok(res)

        now = datetime.now()
        M.date = now.strftime("%Y-%m-%d")
        M.time = now.strftime("%H:%M:%S")

        print('Start')
        start_time = time.time_ns()

        while (time.time_ns() - start_time < duree_ns*1.1) and (stop==False):
            ps2000.ps2000_get_streaming_last_values(
                device.handle,
                callback
            )

        print('Done')

        # if M.fs!=effective_fs/M.upsampling_factor:
        #     M.fs = effective_fs/M.upsampling_factor
        #     print('Warning : Sampling frequency fs changed to nearest possible value of '+str(M.fs)+' Hz')
        #     for i in range(len(M.in_map)):
        #         M.data[M.in_name[i]].fs = M.fs

        # for i in range(len(M.in_map)):
        #     if M.in_map[i] == 1:
        #         if M.upsampling_factor > 1:
        #             M.data[M.in_name[i]].raw = decimate(np.double(adc_to_mv(adc_valuesA, rangeA)[:])/1000,M.upsampling_factor,ftype='fir')[0:int(round(M.dur*M.fs))]
        #         else:
        #             M.data[M.in_name[i]].raw = (np.double(adc_to_mv(adc_valuesA, rangeA)[:])/1000)[0:int(round(M.dur*M.fs))]
        #     elif M.in_map[i] == 2:
        #         if M.upsampling_factor > 1:
        #             M.data[M.in_name[i]].raw = decimate(np.double(adc_to_mv(adc_valuesB, rangeB)[:])/1000,M.upsampling_factor,ftype='fir')[0:int(round(M.dur*M.fs))]
        #         else:
        #             M.data[M.in_name[i]].raw = (np.double(adc_to_mv(adc_valuesB, rangeB)[:])/1000)[0:int(round(M.dur*M.fs))]

def ps4000_plot(M,plotbuffersize=2000,updatetime=0.1):
    """
    This function needs M to contain the following properties:
        - in_range : a list of strings specifying the voltage range.
            Possible voltage ranges are "10MV", "20MV", "50MV", "100MV",
            "200MV", "500MV", "1V", "2V", "5V", "10V", "20V", "50V", "100V"
        - upsampling_factor : upsampling factor
        - in_coupling : Coupling configuration of the channels.
            Can be "ac" or "dc"
    """

    import time


    # Plotting part

    global plotbuffer, ax, x, line, stop, timesincelastupdate, updtime

    updtime = updatetime

    timesincelastupdate = 0.0

    stop = False

    def fstop(event):
        global stop
        stop = True

    def tamp_plus(event):
        ax[0].set_ylim(np.array(ax[0].get_ylim())/2)
    def tamp_moins(event):
        ax[0].set_ylim(np.array(ax[0].get_ylim())*2)
    def famp_plus(event):
        ax[1].set_ylim(np.array(ax[1].get_ylim())/2)
    def famp_moins(event):
        ax[1].set_ylim(np.array(ax[1].get_ylim())*2)
    def on_close(event):
        global stop
        stop = True

    x = np.arange(plotbuffersize)
    plotbuffer = np.sin(x)
    fig, ax = plt.subplots(1,2,figsize=(15,5))
    fig.subplots_adjust(bottom=0.2)
    ax[0].set_xlabel('Temps [s]',fontsize=15)
    ax[0].set_ylabel('Tension [V]',fontsize=15)
    ax[1].set_xlabel('Fréquence [Hz]',fontsize=15)
    ax[1].set_ylabel('Tension [V]',fontsize=15)
    ax[1].set_xlim([0.01,M.fs/2])
    ax[1].set_ylim([0,10])
    linet, = ax[0].plot(x/M.fs,plotbuffer)
    linef, = ax[1].plot(np.fft.fftfreq(n=plotbuffersize,d=1/M.fs),np.fft.fft(plotbuffer,norm='ortho'))
    axs = fig.add_axes([0.4, 0.01, 0.2, 0.075])
    bstop = Button(axs, 'Stop')
    bstop.on_clicked(fstop)
    atxp = fig.add_axes([0.02, 0.6, 0.04, 0.05])
    btplus = Button(atxp, '+')
    btplus.on_clicked(tamp_plus)
    atxm = fig.add_axes([0.02, 0.4, 0.04, 0.05])
    btmoins = Button(atxm, '-')
    btmoins.on_clicked(tamp_moins)
    afxp = fig.add_axes([0.92, 0.6, 0.04, 0.05])
    bfplus = Button(afxp, '+')
    bfplus.on_clicked(famp_plus)
    afxm = fig.add_axes([0.92, 0.4, 0.04, 0.05])
    bfmoins = Button(afxm, '-')
    bfmoins.on_clicked(famp_moins)
    fig.canvas.mpl_connect('close_event', on_close)
    plt.pause(0.0001)

    # End of plotting part

    # Find maximum ADC count value
    # handle = chandle
    # pointer to value = ctypes.byref(maxADC)
    maxADC = ctypes.c_int16(32767)

    global nextSample, autoStopOuter, wasCalledBack

    # Buffer size fixed to 20k samples
    sizeOfOneBuffer = 20_000

    # Effective sampling frequency
    # If upsampling is > 1, the actual data acquisition is
    # performed at fs*upsampling, and then decimated to
    # the desired frequency
    effective_fs = M.fs * M.upsampling_factor

    # Sample interval is the duration between two consecutive samples
    # As it is the sampling frequency that is specified,
    # and the sampling interval can only take integer increments
    # of 1 nanoseconds, the actual sampling frequency might
    # necessitate adjustments
    si = round(1e9/effective_fs)
    if effective_fs != (1e9/si):
        effective_fs = (1e9/si)
        print("Warning : Sampling frequency fs changed to nearest possible value of "+str(effective_fs)+" Hz")
    
    print("Effective sampling frequency: "+str(effective_fs))

    numdesiredsamples = int(round(effective_fs*M.dur))
    numBuffersToCapture = int(np.ceil(numdesiredsamples/sizeOfOneBuffer))
    sampleInterval = ctypes.c_int32(si)
    totalSamples = sizeOfOneBuffer * numBuffersToCapture
    chandle = ctypes.c_int16()
    status = {}

    # Open PicoScope 4000 Series device
    # Returns handle to chandle for use in future API functions
    status["openunit"] = ps4000.ps4000OpenUnit(ctypes.byref(chandle))
    assert_pico_ok(status["openunit"])

    # Setup channel A
    indA = findindex(M.in_map,1)
    if indA!=None:
        enabledA = True
        rangeA = ps4000.PS4000_RANGE['PS4000_'+M.in_range[indA]]
        if M.in_coupling[indA].capitalize()=='Dc':
           couplingA = 1
        elif M.in_coupling[indA].capitalize()=='Ac':
            couplingA = 0
        else:
            print("Input A coupling not recognized, set to 'dc'")
            couplingA = 1
            M.in_coupling[indA]='dc'
        # Create buffers ready for assigning pointers for data collection
        bufferAMax = np.zeros(shape=sizeOfOneBuffer, dtype=np.int16)
        # Total buffer
        bufferCompleteA = np.zeros(shape=totalSamples, dtype=np.int16)
        print('Channel A: enabled with range '+'PS4000_'+M.in_range[indA]+' ('+str(rangeA)+')')
    else:
        enabledA = False
        rangeA = ps4000.PS4000_RANGE['PS4000_10V']
        print('Channel A: disabled')

    # Set up channel A
    channel_range = ps4000.PS4000_RANGE['PS4000_50MV']
    status["setChA"] = ps4000.ps4000SetChannel(chandle,
                                            ps4000.PS4000_CHANNEL['PS4000_CHANNEL_A'],
                                            int(enabledA),
                                            couplingA,
                                            rangeA)
    assert_pico_ok(status["setChA"])

    # Setup channel B
    indB = findindex(M.in_map,2)
    if indB!=None:
        enabledB = True
        rangeB = ps4000.PS4000_RANGE['PS4000_'+M.in_range[indB]]
        # Create buffers ready for assigning pointers for data collection
        bufferBMax = np.zeros(shape=sizeOfOneBuffer, dtype=np.int16)
        # Total buffer
        bufferCompleteB = np.zeros(shape=totalSamples, dtype=np.int16)
        print('Channel B: enabled with range '+'PS4000_'+M.in_range[indB]+' ('+str(rangeB)+')')
        if M.in_coupling[indB].capitalize()=='Dc':
           couplingB = 1
        elif M.in_coupling[indB].capitalize()=='Ac':
            couplingB = 0
        else:
            print("Input B coupling not recognized, set to 'dc'")
            couplingB = 1
            M.in_coupling[indB]='dc'
    else:
        enabledB = False
        couplingB = 1
        rangeB = ps4000.PS4000_RANGE['PS4000_10V']
        print('Channel B: disabled')

    # Set up channel B
    status["setChB"] = ps4000.ps4000SetChannel(chandle,
                                            ps4000.PS4000_CHANNEL['PS4000_CHANNEL_B'],
                                            int(enabledB),
                                            couplingB,
                                            rangeB)
    assert_pico_ok(status["setChB"])

    # Set data buffer location for data collection from channel A
    # Parameters :  handle = chandle
    #               source = PS4000_CHANNEL_A = 0
    #               pointer to buffer max = ctypes.byref(bufferAMax)
    #               pointer to buffer min = ctypes.byref(bufferAMin)
    #               buffer length = maxSamples
    #               segment index = 0
    #               ratio mode = PS4000_RATIO_MODE_NONE = 0
    if enabledA:
        status["setDataBuffersA"] = ps4000.ps4000SetDataBuffers(chandle,
                                                        ps4000.PS4000_CHANNEL['PS4000_CHANNEL_A'],
                                                        bufferAMax.ctypes.data_as(ctypes.POINTER(ctypes.c_int16)),
                                                        None,
                                                        sizeOfOneBuffer)
        assert_pico_ok(status["setDataBuffersA"])

    # Set data buffer location for data collection from channel B
    # Parameters :  handle = chandle
    #               source = PS4000_CHANNEL_B = 0
    #               pointer to buffer max = ctypes.byref(bufferAMax)
    #               pointer to buffer min = ctypes.byref(bufferAMin)
    #               buffer length = maxSamples
    #               segment index = 0
    #               ratio mode = PS4000_RATIO_MODE_NONE = 0
    if enabledB:
        status["setDataBuffersB"] = ps4000.ps4000SetDataBuffers(chandle,
                                                        ps4000.PS4000_CHANNEL['PS4000_CHANNEL_B'],
                                                        bufferBMax.ctypes.data_as(ctypes.POINTER(ctypes.c_int16)),
                                                        None,
                                                        sizeOfOneBuffer)
        assert_pico_ok(status["setDataBuffersB"])

    now = datetime.now()
    M.date = now.strftime("%Y-%m-%d")
    M.time = now.strftime("%H:%M:%S")

    # Begin streaming mode:
    sampleUnits = ps4000.PS4000_TIME_UNITS['PS4000_NS']
    # We are not triggering:
    maxPreTriggerSamples = 0
    autoStopOn = 1
    # No downsampling:
    downsampleRatio = 1
    status["runStreaming"] = ps4000.ps4000RunStreaming(chandle,
                                                    ctypes.byref(sampleInterval),
                                                    sampleUnits,
                                                    maxPreTriggerSamples,
                                                    totalSamples,
                                                    autoStopOn,
                                                    downsampleRatio,
                                                    sizeOfOneBuffer)
    assert_pico_ok(status["runStreaming"])

    actualSampleInterval = sampleInterval.value
    actualSampleIntervalNs = actualSampleInterval * 1000

    print("Capturing at sample interval %s ns" % actualSampleIntervalNs)

    nextSample = 0
    autoStopOuter = False
    wasCalledBack = False

    def streaming_callback(handle, noOfSamples, startIndex, overflow, triggerAt, triggered, autoStop, param):

        global plotbuffer, timesincelastupdate, updtime
        global nextSample, autoStopOuter, wasCalledBack
        wasCalledBack = True
        destEnd = nextSample + noOfSamples
        sourceEnd = startIndex + noOfSamples
        timesincelastupdate += noOfSamples/M.fs
        plotbuffer=np.roll(plotbuffer,int(-noOfSamples),axis=0)
        plotbuffer[-noOfSamples:] = (np.array(adc2mV(bufferAMax[startIndex:sourceEnd],rangeA,maxADC))/1000)
        if timesincelastupdate > updtime:
            linet.set_ydata(plotbuffer)
            linef.set_ydata(np.abs(np.fft.fft(plotbuffer*np.hanning(plotbuffersize),norm='ortho')))
            plt.pause(0.0001)
            timesincelastupdate = 0.0
        # if enabledA:
        #     bufferCompleteA[nextSample:destEnd] = bufferAMax[startIndex:sourceEnd]
        # if enabledB:
        #     bufferCompleteB[nextSample:destEnd] = bufferBMax[startIndex:sourceEnd]
        nextSample += noOfSamples
        if autoStop:
            autoStopOuter = True

    # Convert the python function into a C function pointer.
    cFuncPtr = ps4000.StreamingReadyType(streaming_callback)

    # Fetch data from the driver in a loop, copying it out of the registered buffers and into our complete one.
    while nextSample < totalSamples and not autoStopOuter and not stop:
        wasCalledBack = False
        status["getStreamingLastestValues"] = ps4000.ps4000GetStreamingLatestValues(chandle, cFuncPtr, None)
        if not wasCalledBack:
            # If we weren't called back by the driver, this means no data is ready. Sleep for a short while before trying
            # again.
            time.sleep(0.01)

    print("Done grabbing values.")

    # # Convert ADC counts data to mV
    # if enabledA:
    #     adc2mVChAMax = adc2mV(bufferCompleteA, rangeA, maxADC)
    # if enabledB:
    #     adc2mVChBMax = adc2mV(bufferCompleteB, rangeB, maxADC)

    # # Create time data
    # time = np.linspace(0, (totalSamples - 1) * actualSampleIntervalNs, totalSamples)

    # Stop the scope
    # handle = chandle
    status["stop"] = ps4000.ps4000Stop(chandle)
    assert_pico_ok(status["stop"])

    # Disconnect the scope
    # handle = chandle
    status["close"] = ps4000.ps4000CloseUnit(chandle)
    assert_pico_ok(status["close"])

    # for i in range(len(M.in_map)):
    #     if M.in_map[i] == 1:
    #         M.data[M.in_name[i]].raw = decimate(np.double(adc2mVChAMax[0:round(M.dur*effective_fs)])/1000,M.upsampling_factor)
    #     elif M.in_map[i] == 2:
    #         M.data[M.in_name[i]].raw = decimate(np.double(adc2mVChBMax[0:round(M.dur*effective_fs)])/1000,M.upsampling_factor)

    # if M.fs!=effective_fs/M.upsampling_factor:
    #     M.fs = effective_fs/M.upsampling_factor
    #     print('Warning : Sampling frequency fs changed to nearest possible value of '+str(M.fs)+' Hz')
    #     for i in range(len(M.in_map)):
    #         M.data[M.in_name[i]].fs = M.fs
    

def ps4000_run_measurement(M):
    """
    This function needs M to contain the following properties:
        - in_range : a list of strings specifying the voltage range.
            Possible voltage ranges are "10MV", "20MV", "50MV", "100MV",
            "200MV", "500MV", "1V", "2V", "5V", "10V", "20V", "50V", "100V"
        - upsampling_factor : upsampling factor
        - in_coupling : Coupling configuration of the channels.
            Can be "ac" or "dc"
    """

    import time
    global nextSample, autoStopOuter, wasCalledBack

    # Buffer size fixed to 20k samples
    sizeOfOneBuffer = 100_000

    # Effective sampling frequency
    # If upsampling is > 1, the actual data acquisition is
    # performed at fs*upsampling, and then decimated to
    # the desired frequency
    effective_fs = M.fs * M.upsampling_factor

    # Sample interval is the duration between two consecutive samples
    # As it is the sampling frequency that is specified,
    # and the sampling interval can only take integer increments
    # of 1 nanoseconds, the actual sampling frequency might
    # necessitate adjustments
    si = round(1e9/effective_fs)
    if effective_fs != (1e9/si):
        effective_fs = (1e9/si)
        print("Warning : Sampling frequency fs changed to nearest possible value of "+str(effective_fs)+" Hz")
    
    print("Effective sampling frequency: "+str(effective_fs))

    numdesiredsamples = int(round(effective_fs*M.dur))
    numBuffersToCapture = int(np.ceil(numdesiredsamples/sizeOfOneBuffer))
    sampleInterval = ctypes.c_int32(si)
    totalSamples = sizeOfOneBuffer * numBuffersToCapture
    chandle = ctypes.c_int16()
    status = {}

    # Open PicoScope 4000 Series device
    # Returns handle to chandle for use in future API functions
    status["openunit"] = ps4000.ps4000OpenUnit(ctypes.byref(chandle))
    assert_pico_ok(status["openunit"])

    # Setup channel A
    indA = findindex(M.in_map,1)
    if indA!=None:
        enabledA = True
        rangeA = ps4000.PS4000_RANGE['PS4000_'+M.in_range[indA]]
        if M.in_coupling[indA].capitalize()=='Dc':
           couplingA = 1
        elif M.in_coupling[indA].capitalize()=='Ac':
            couplingA = 0
        else:
            print("Input A coupling not recognized, set to 'dc'")
            couplingA = 1
            M.in_coupling[indA]='dc'
        # Create buffers ready for assigning pointers for data collection
        bufferAMax = np.zeros(shape=sizeOfOneBuffer, dtype=np.int16)
        # Total buffer
        bufferCompleteA = np.zeros(shape=totalSamples, dtype=np.int16)
        print('Channel A: enabled with range '+'PS4000_'+M.in_range[indA]+' ('+str(rangeA)+')')
    else:
        enabledA = False
        rangeA = ps4000.PS4000_RANGE['PS4000_10V']
        print('Channel A: disabled')

    # Set up channel A
    channel_range = ps4000.PS4000_RANGE['PS4000_50MV']
    status["setChA"] = ps4000.ps4000SetChannel(chandle,
                                            ps4000.PS4000_CHANNEL['PS4000_CHANNEL_A'],
                                            int(enabledA),
                                            couplingA,
                                            rangeA)
    assert_pico_ok(status["setChA"])

    # Setup channel B
    indB = findindex(M.in_map,2)
    if indB!=None:
        enabledB = True
        rangeB = ps4000.PS4000_RANGE['PS4000_'+M.in_range[indB]]
        # Create buffers ready for assigning pointers for data collection
        bufferBMax = np.zeros(shape=sizeOfOneBuffer, dtype=np.int16)
        # Total buffer
        bufferCompleteB = np.zeros(shape=totalSamples, dtype=np.int16)
        print('Channel B: enabled with range '+'PS4000_'+M.in_range[indB]+' ('+str(rangeB)+')')
        if M.in_coupling[indB].capitalize()=='Dc':
           couplingB = 1
        elif M.in_coupling[indB].capitalize()=='Ac':
            couplingB = 0
        else:
            print("Input B coupling not recognized, set to 'dc'")
            couplingB = 1
            M.in_coupling[indB]='dc'
    else:
        enabledB = False
        rangeB = ps4000.PS4000_RANGE['PS4000_10V']
        print('Channel B: disabled')

    # Set up channel B
    status["setChB"] = ps4000.ps4000SetChannel(chandle,
                                            ps4000.PS4000_CHANNEL['PS4000_CHANNEL_B'],
                                            int(enabledB),
                                            couplingB,
                                            rangeB)
    assert_pico_ok(status["setChB"])

    # Set data buffer location for data collection from channel A
    # Parameters :  handle = chandle
    #               source = PS4000_CHANNEL_A = 0
    #               pointer to buffer max = ctypes.byref(bufferAMax)
    #               pointer to buffer min = ctypes.byref(bufferAMin)
    #               buffer length = maxSamples
    #               segment index = 0
    #               ratio mode = PS4000_RATIO_MODE_NONE = 0
    if enabledA:
        status["setDataBuffersA"] = ps4000.ps4000SetDataBuffers(chandle,
                                                        ps4000.PS4000_CHANNEL['PS4000_CHANNEL_A'],
                                                        bufferAMax.ctypes.data_as(ctypes.POINTER(ctypes.c_int16)),
                                                        None,
                                                        sizeOfOneBuffer)
        assert_pico_ok(status["setDataBuffersA"])

    # Set data buffer location for data collection from channel B
    # Parameters :  handle = chandle
    #               source = PS4000_CHANNEL_B = 0
    #               pointer to buffer max = ctypes.byref(bufferAMax)
    #               pointer to buffer min = ctypes.byref(bufferAMin)
    #               buffer length = maxSamples
    #               segment index = 0
    #               ratio mode = PS4000_RATIO_MODE_NONE = 0
    if enabledB:
        status["setDataBuffersB"] = ps4000.ps4000SetDataBuffers(chandle,
                                                        ps4000.PS4000_CHANNEL['PS4000_CHANNEL_B'],
                                                        bufferBMax.ctypes.data_as(ctypes.POINTER(ctypes.c_int16)),
                                                        None,
                                                        sizeOfOneBuffer)
        assert_pico_ok(status["setDataBuffersB"])

    now = datetime.now()
    M.date = now.strftime("%Y-%m-%d")
    M.time = now.strftime("%H:%M:%S")

    # Begin streaming mode:
    sampleUnits = ps4000.PS4000_TIME_UNITS['PS4000_NS']
    # We are not triggering:
    maxPreTriggerSamples = 0
    autoStopOn = 1
    # No downsampling:
    downsampleRatio = 1
    status["runStreaming"] = ps4000.ps4000RunStreaming(chandle,
                                                    ctypes.byref(sampleInterval),
                                                    sampleUnits,
                                                    maxPreTriggerSamples,
                                                    totalSamples,
                                                    autoStopOn,
                                                    downsampleRatio,
                                                    sizeOfOneBuffer)
    assert_pico_ok(status["runStreaming"])

    actualSampleInterval = sampleInterval.value
    actualSampleIntervalNs = actualSampleInterval * 1000

    print("Capturing at sample interval %s ns" % actualSampleIntervalNs)

    nextSample = 0
    autoStopOuter = False
    wasCalledBack = False

    def streaming_callback(handle, noOfSamples, startIndex, overflow, triggerAt, triggered, autoStop, param):
        global nextSample, autoStopOuter, wasCalledBack
        wasCalledBack = True
        destEnd = nextSample + noOfSamples
        sourceEnd = startIndex + noOfSamples
        if enabledA:
            bufferCompleteA[nextSample:destEnd] = bufferAMax[startIndex:sourceEnd]
        if enabledB:
            bufferCompleteB[nextSample:destEnd] = bufferBMax[startIndex:sourceEnd]
        nextSample += noOfSamples
        if autoStop:
            autoStopOuter = True


    # Convert the python function into a C function pointer.
    cFuncPtr = ps4000.StreamingReadyType(streaming_callback)

    # Fetch data from the driver in a loop, copying it out of the registered buffers and into our complete one.
    while nextSample < totalSamples and not autoStopOuter:
        wasCalledBack = False
        status["getStreamingLastestValues"] = ps4000.ps4000GetStreamingLatestValues(chandle, cFuncPtr, None)
        if not wasCalledBack:
            # If we weren't called back by the driver, this means no data is ready. Sleep for a short while before trying
            # again.
            time.sleep(0.01)

    print("Done grabbing values.")

    # Find maximum ADC count value
    # handle = chandle
    # pointer to value = ctypes.byref(maxADC)
    maxADC = ctypes.c_int16(32767)

    # Convert ADC counts data to mV
    if enabledA:
        adc2mVChAMax = adc2mV(bufferCompleteA, rangeA, maxADC)
    if enabledB:
        adc2mVChBMax = adc2mV(bufferCompleteB, rangeB, maxADC)

    # Create time data
    time = np.linspace(0, (totalSamples - 1) * actualSampleIntervalNs, totalSamples)

    # Stop the scope
    # handle = chandle
    status["stop"] = ps4000.ps4000Stop(chandle)
    assert_pico_ok(status["stop"])

    # Disconnect the scope
    # handle = chandle
    status["close"] = ps4000.ps4000CloseUnit(chandle)
    assert_pico_ok(status["close"])

    for i in range(len(M.in_map)):
        if M.in_map[i] == 1:
            M.data[M.in_name[i]].raw = decimate(np.double(adc2mVChAMax[0:round(M.dur*effective_fs)])/1000,M.upsampling_factor)
        elif M.in_map[i] == 2:
            M.data[M.in_name[i]].raw = decimate(np.double(adc2mVChBMax[0:round(M.dur*effective_fs)])/1000,M.upsampling_factor)

    if M.fs!=effective_fs/M.upsampling_factor:
        M.fs = effective_fs/M.upsampling_factor
        print('Warning : Sampling frequency fs changed to nearest possible value of '+str(M.fs)+' Hz')
        for i in range(len(M.in_map)):
            M.data[M.in_name[i]].fs = M.fs
    


# _tools.py
#
# Various tools for measpy
#
# (c) OD, 2021

import csv
import numpy as np

def csv_to_dict(filename):
    """ Conversion from a CSV (produced by the class Measurement) to a dict
          Default separator is (,)
          First row is the key string
          The value is a list
    """
    dd={}
    with open(filename, 'r') as file:
        reader = csv.reader(file)
        for row in reader:
            dd[row[0]]=row[1:]
    return dd

def convl(fun,xx):
    if type(xx)==list:
        yy=list(map(fun,xx))
    else:
        yy=fun(xx) 
    return yy

def convl1(fun,xx):
    if type(xx)==list:
        yy=fun(xx[0])
    else:
        yy=fun(xx) 
    return yy

def add_step(a,b):
    return a+'\n -->'+b

def wrap(phase):
    """ Opposite of np.unwrap   
    """
    return np.mod((phase + np.pi), (2 * np.pi)) - np.pi

def unwrap_around_index(phase,n):
    """ Opposite of np.unwrap   
    """
    return np.hstack((np.unwrap(phase[n-1::-1])[::-1],np.unwrap(phase[n:])))

def smooth(in_array,l=20):
    ker = np.ones(l)/l
    return np.convolve(in_array,ker,mode='same')

def nth_octave_bands(n,fmin=5,fmax=20000):
    """ 1/nth octave band frequency range calculation """
    nmin = int(np.ceil(n*np.log2(fmin*10**-3)))
    nmax = int(np.ceil(n*np.log2(fmax*10**-3)))
    indices = range(nmin,nmax+1)
    f_centre = 1000 * (2**(np.array(indices)/n))
    f2 = 2**(1/n/2)
    f_upper = f_centre * f2
    f_lower = f_centre / f2
    return f_centre, f_lower, f_upper

def picv(long):
    """ Create a 1D-array of length long with a unitary peak in the middle """
    return np.hstack((np.zeros(long),1,np.zeros(long-1)))

def _create_time1(fs,dur):
    return np.linspace(0,dur-1/fs,int(round(dur*fs)))  # time axis

def _create_time2(fs,length):
    return np.linspace(0,(length-1)/fs,length)  # time axis

def create_time(fs,dur=None,length=None):
    if dur==None and length==None:
        raise Exception('dur=duration in s or length=number of samples must be specified.')
    if dur!=None and length!=None:
        raise Exception("dur and length can't be both specified.")
    if dur!=None:
        return _create_time1(fs,dur)
    else:
        return _create_time2(fs,length)

def apply_fades(s,fades):
    if fades[0]>0:
        s[0:fades[0]] = s[0:fades[0]] * ((-np.cos(np.arange(fades[0])/fades[0]*np.pi)+1) / 2)
    if fades[1]>0:
        s[-fades[1]:] = s[-fades[1]:] *  ((np.cos(np.arange(fades[1])/fades[1]*np.pi)+1) / 2)
    return s
